@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2015, Sonos, Inc.
+ * Copyright (c) 2015-2017, Sonos, Inc.
  *
  * SPDX-License-Identifier:	GPL-2.0
  *
@@ -39,8 +39,7 @@ sonosHash(SonosDigestAlg_t alg, const void *buf, size_t bufLen,
 		return 0;
 	}
 
-	sg_init_table(&sg, 1);
-	sg_set_buf(&sg, buf, bufLen);
+	sg_init_one(&sg, buf, bufLen);
 	desc.tfm = tfm;
 	desc.flags = 0;
 
