@@ -60,13 +60,13 @@ static int get_exclusive(struct ubi_volume_desc *desc)
 	struct ubi_volume *vol = desc->vol;
 
 	spin_lock(&vol->ubi->volumes_lock);
-	users = vol->readers + vol->writers + vol->exclusive;
+	users = vol->readers + vol->writers + vol->exclusive + vol->metaonly;
 	ubi_assert(users > 0);
 	if (users > 1) {
-		ubi_err("%d users for volume %d", users, vol->vol_id);
+		ubi_err(vol->ubi, "%d users for volume %d", users, vol->vol_id);
 		err = -EBUSY;
 	} else {
-		vol->readers = vol->writers = 0;
+		vol->readers = vol->writers = vol->metaonly = 0;
 		vol->exclusive = 1;
 		err = desc->mode;
 		desc->mode = UBI_EXCLUSIVE;
@@ -86,13 +86,15 @@ static void revoke_exclusive(struct ubi_volume_desc *desc, int mode)
 	struct ubi_volume *vol = desc->vol;
 
 	spin_lock(&vol->ubi->volumes_lock);
-	ubi_assert(vol->readers == 0 && vol->writers == 0);
+	ubi_assert(vol->readers == 0 && vol->writers == 0 && vol->metaonly == 0);
 	ubi_assert(vol->exclusive == 1 && desc->mode == UBI_EXCLUSIVE);
 	vol->exclusive = 0;
 	if (mode == UBI_READONLY)
 		vol->readers = 1;
 	else if (mode == UBI_READWRITE)
 		vol->writers = 1;
+	else if (mode == UBI_METAONLY)
+		vol->metaonly = 1;
 	else
 		vol->exclusive = 1;
 	spin_unlock(&vol->ubi->volumes_lock);
@@ -134,7 +136,7 @@ static int vol_cdev_release(struct inode *inode, struct file *file)
 		vol->ubi->ubi_num, vol->vol_id, desc->mode);
 
 	if (vol->updating) {
-		ubi_warn("update of volume %d not finished, volume is damaged",
+		ubi_warn(vol->ubi, "update of volume %d not finished, volume is damaged",
 			 vol->vol_id);
 		ubi_assert(!vol->changing_leb);
 		vol->updating = 0;
@@ -155,38 +157,14 @@ static loff_t vol_cdev_llseek(struct file *file, loff_t offset, int origin)
 {
 	struct ubi_volume_desc *desc = file->private_data;
 	struct ubi_volume *vol = desc->vol;
-	loff_t new_offset;
 
 	if (vol->updating) {
 		/* Update is in progress, seeking is prohibited */
-		ubi_err("updating");
+		ubi_err(vol->ubi, "updating");
 		return -EBUSY;
 	}
 
-	switch (origin) {
-	case 0: /* SEEK_SET */
-		new_offset = offset;
-		break;
-	case 1: /* SEEK_CUR */
-		new_offset = file->f_pos + offset;
-		break;
-	case 2: /* SEEK_END */
-		new_offset = vol->used_bytes + offset;
-		break;
-	default:
-		return -EINVAL;
-	}
-
-	if (new_offset < 0 || new_offset > vol->used_bytes) {
-		ubi_err("bad seek %lld", new_offset);
-		return -EINVAL;
-	}
-
-	dbg_gen("seek volume %d, offset %lld, origin %d, new offset %lld",
-		vol->vol_id, offset, origin, new_offset);
-
-	file->f_pos = new_offset;
-	return new_offset;
+	return fixed_size_llseek(file, offset, origin, vol->used_bytes);
 }
 
 static int vol_cdev_fsync(struct file *file, loff_t start, loff_t end,
@@ -217,11 +195,11 @@ static ssize_t vol_cdev_read(struct file *file, __user char *buf, size_t count,
 		count, *offp, vol->vol_id);
 
 	if (vol->updating) {
-		ubi_err("updating");
+		ubi_err(vol->ubi, "updating");
 		return -EBUSY;
 	}
 	if (vol->upd_marker) {
-		ubi_err("damaged volume, update marker is set");
+		ubi_err(vol->ubi, "damaged volume, update marker is set");
 		return -EBADF;
 	}
 	if (*offp == vol->used_bytes || count == 0)
@@ -301,7 +279,7 @@ static ssize_t vol_cdev_direct_write(struct file *file, const char __user *buf,
 
 	lnum = div_u64_rem(*offp, vol->usable_leb_size, &off);
 	if (off & (ubi->min_io_size - 1)) {
-		ubi_err("unaligned position");
+		ubi_err(ubi, "unaligned position");
 		return -EINVAL;
 	}
 
@@ -310,7 +288,7 @@ static ssize_t vol_cdev_direct_write(struct file *file, const char __user *buf,
 
 	/* We can write only in fractions of the minimum I/O unit */
 	if (count & (ubi->min_io_size - 1)) {
-		ubi_err("unaligned write length");
+		ubi_err(ubi, "unaligned write length");
 		return -EINVAL;
 	}
 
@@ -372,7 +350,7 @@ static ssize_t vol_cdev_write(struct file *file, const char __user *buf,
 		err = ubi_more_leb_change_data(ubi, vol, buf, count);
 
 	if (err < 0) {
-		ubi_err("cannot accept more %zd bytes of data, error %d",
+		ubi_err(ubi, "cannot accept more %zd bytes of data, error %d",
 			count, err);
 		return err;
 	}
@@ -394,7 +372,7 @@ static ssize_t vol_cdev_write(struct file *file, const char __user *buf,
 			return err;
 
 		if (err) {
-			ubi_warn("volume %d on UBI device %d is corrupted",
+			ubi_warn(ubi, "volume %d on UBI device %d is corrupted",
 				 vol->vol_id, ubi->ubi_num);
 			vol->corrupted = 1;
 		}
@@ -406,127 +384,6 @@ static ssize_t vol_cdev_write(struct file *file, const char __user *buf,
 	return count;
 }
 
-#ifdef CONFIG_SONOS
-struct ubi_volume_desc * rootfs_open(int vol_id, int ubi_num)
-{
-	struct ubi_volume_desc *desc;
-
-	desc = ubi_open_volume(ubi_num, vol_id, UBI_READWRITE);
-	if (IS_ERR(desc)) {
-		printk("rootfs open device %d, volume %d fail %ld\n",
-			ubi_num, vol_id, PTR_ERR(desc));
-		return NULL;
-	}
-	return desc;
-}
-
-int rootfs_update(struct ubi_volume_desc *desc, int64_t bytes)
-{
-	int err = 0;
-	struct ubi_volume *vol = desc->vol;
-	struct ubi_device *ubi = vol->ubi;
-	int64_t rsvd_bytes;
-
-	if (!capable(CAP_SYS_RESOURCE)) {
-		err = -EPERM;
-		goto  err_skip;
-	}
-
-	if (desc->mode == UBI_READONLY) {
-		err = -EROFS;
-		goto err_skip;
-	}
-
-	rsvd_bytes = (long long)vol->reserved_pebs *
-			ubi->leb_size-vol->data_pad;
-	if (bytes < 0 || bytes > rsvd_bytes) {
-		err = -EINVAL;
-		goto err_skip;
-	}
-
-	err = get_exclusive(desc);
-	if (err < 0)
-		goto err_skip;
-
-	err = ubi_start_update(ubi, vol, bytes);
-	if (bytes == 0)
-		revoke_exclusive(desc, UBI_READWRITE);
-err_skip:
-	return err;
-}
-
-ssize_t rootfs_write(struct ubi_volume_desc *desc, const char *buf,
-			      size_t count, int flag)
-{
-	int err = 0;
-	struct ubi_volume *vol = desc->vol;
-	struct ubi_device *ubi = vol->ubi;
-
-	if ( !vol->updating ) {
-		printk("rootfs_write not ready\n");
-		return -1;
-	}
-
-	err = rootfs_ubi_more_update_data(ubi, vol, buf, count, flag);
-
-	if (err < 0) {
-		ubi_err("cannot accept more %zd bytes of data, error %d",
-			count, err);
-		return err;
-	}
-
-	if (err) {
-		/*
-		 * The operation is finished, @err contains number of actually
-		 * written bytes.
-		 */
-		count = err;
-
-		if (vol->changing_leb) {
-			revoke_exclusive(desc, UBI_READWRITE);
-			return count;
-		}
-
-		err = ubi_check_volume(ubi, vol->vol_id);
-		if (err < 0)
-			return err;
-
-		if (err) {
-			ubi_warn("volume %d on UBI device %d is corrupted",
-				 vol->vol_id, ubi->ubi_num);
-			vol->corrupted = 1;
-		}
-		vol->checked = 1;
-		ubi_volume_notify(ubi, vol, UBI_VOLUME_UPDATED);
-		revoke_exclusive(desc, UBI_READWRITE);
-	}
-
-	return count;
-}
-
-int rootfs_release(struct ubi_volume_desc *desc)
-{
-	struct ubi_volume *vol = desc->vol;
-
-	if (vol->updating) {
-		ubi_warn("update of volume %d not finished, volume is damaged",
-			 vol->vol_id);
-		ubi_assert(!vol->changing_leb);
-		vol->updating = 0;
-		vfree(vol->upd_buf);
-	} else if (vol->changing_leb) {
-		dbg_gen("only %lld of %lld bytes received for atomic LEB change for volume %d:%d, cancel",
-			vol->upd_received, vol->upd_bytes, vol->ubi->ubi_num,
-			vol->vol_id);
-		vol->changing_leb = 0;
-		vfree(vol->upd_buf);
-	}
-
-	ubi_close_volume(desc);
-	return 0;
-}
-#endif
-
 static long vol_cdev_ioctl(struct file *file, unsigned int cmd,
 			   unsigned long arg)
 {
@@ -570,8 +427,10 @@ static long vol_cdev_ioctl(struct file *file, unsigned int cmd,
 			break;
 
 		err = ubi_start_update(ubi, vol, bytes);
-		if (bytes == 0)
+		if (bytes == 0) {
+			ubi_volume_notify(ubi, vol, UBI_VOLUME_UPDATED);
 			revoke_exclusive(desc, UBI_READWRITE);
+		}
 		break;
 	}
 
@@ -596,7 +455,7 @@ static long vol_cdev_ioctl(struct file *file, unsigned int cmd,
 		/* Validate the request */
 		err = -EINVAL;
 		if (req.lnum < 0 || req.lnum >= vol->reserved_pebs ||
-		    req.bytes < 0 || req.lnum >= vol->usable_leb_size)
+		    req.bytes < 0 || req.bytes > vol->usable_leb_size)
 			break;
 
 		err = get_exclusive(desc);
@@ -706,6 +565,26 @@ static long vol_cdev_ioctl(struct file *file, unsigned int cmd,
 		break;
 	}
 
+	/* Create a R/O block device on top of the UBI volume */
+	case UBI_IOCVOLCRBLK:
+	{
+		struct ubi_volume_info vi;
+
+		ubi_get_volume_info(desc, &vi);
+		err = ubiblock_create(&vi);
+		break;
+	}
+
+	/* Remove the R/O block device */
+	case UBI_IOCVOLRMBLK:
+	{
+		struct ubi_volume_info vi;
+
+		ubi_get_volume_info(desc, &vi);
+		err = ubiblock_remove(&vi);
+		break;
+	}
+
 	default:
 		err = -ENOTTY;
 		break;
@@ -765,7 +644,7 @@ static int verify_mkvol_req(const struct ubi_device *ubi,
 	return 0;
 
 bad:
-	ubi_err("bad volume creation request");
+	ubi_err(ubi, "bad volume creation request");
 	ubi_dump_mkvol_req(req);
 	return err;
 }
@@ -842,19 +721,19 @@ static int rename_volumes(struct ubi_device *ubi,
 		req->ents[i].name[req->ents[i].name_len] = '\0';
 		n = strlen(req->ents[i].name);
 		if (n != req->ents[i].name_len)
-			err = -EINVAL;
+			return -EINVAL;
 	}
 
 	/* Make sure volume IDs and names are unique */
 	for (i = 0; i < req->count - 1; i++) {
 		for (n = i + 1; n < req->count; n++) {
 			if (req->ents[i].vol_id == req->ents[n].vol_id) {
-				ubi_err("duplicated volume id %d",
+				ubi_err(ubi, "duplicated volume id %d",
 					req->ents[i].vol_id);
 				return -EINVAL;
 			}
 			if (!strcmp(req->ents[i].name, req->ents[n].name)) {
-				ubi_err("duplicated volume name \"%s\"",
+				ubi_err(ubi, "duplicated volume name \"%s\"",
 					req->ents[i].name);
 				return -EINVAL;
 			}
@@ -874,10 +753,11 @@ static int rename_volumes(struct ubi_device *ubi,
 			goto out_free;
 		}
 
-		re->desc = ubi_open_volume(ubi->ubi_num, vol_id, UBI_EXCLUSIVE);
+		re->desc = ubi_open_volume(ubi->ubi_num, vol_id, UBI_METAONLY);
 		if (IS_ERR(re->desc)) {
 			err = PTR_ERR(re->desc);
-			ubi_err("cannot open volume %d, error %d", vol_id, err);
+			ubi_err(ubi, "cannot open volume %d, error %d",
+				vol_id, err);
 			kfree(re);
 			goto out_free;
 		}
@@ -936,7 +816,7 @@ static int rename_volumes(struct ubi_device *ubi,
 				continue;
 
 			/* The volume exists but busy, or an error occurred */
-			ubi_err("cannot open volume \"%s\", error %d",
+			ubi_err(ubi, "cannot open volume \"%s\", error %d",
 				re->new_name, err);
 			goto out_free;
 		}
@@ -1087,7 +967,7 @@ static long ubi_cdev_ioctl(struct file *file, unsigned int cmd,
 		if (!req) {
 			err = -ENOMEM;
 			break;
-		};
+		}
 
 		err = copy_from_user(req, argp, sizeof(struct ubi_rnvol_req));
 		if (err) {
@@ -1290,3 +1170,124 @@ const struct file_operations ubi_ctrl_cdev_operations = {
 	.compat_ioctl   = ctrl_cdev_compat_ioctl,
 	.llseek		= no_llseek,
 };
+
+#ifdef CONFIG_SONOS
+struct ubi_volume_desc * rootfs_open(int vol_id, int ubi_num)
+{
+	struct ubi_volume_desc *desc;
+
+	desc = ubi_open_volume(ubi_num, vol_id, UBI_READWRITE);
+	if (IS_ERR(desc)) {
+		printk("rootfs open device %d, volume %d fail %ld\n",
+			ubi_num, vol_id, PTR_ERR(desc));
+		return NULL;
+	}
+	return desc;
+}
+
+int rootfs_update(struct ubi_volume_desc *desc, int64_t bytes)
+{
+	int err = 0;
+	struct ubi_volume *vol = desc->vol;
+	struct ubi_device *ubi = vol->ubi;
+	int64_t rsvd_bytes;
+
+	if (!capable(CAP_SYS_RESOURCE)) {
+		err = -EPERM;
+		goto  err_skip;
+	}
+
+	if (desc->mode == UBI_READONLY) {
+		err = -EROFS;
+		goto err_skip;
+	}
+
+	rsvd_bytes = (long long)vol->reserved_pebs *
+			ubi->leb_size-vol->data_pad;
+	if (bytes < 0 || bytes > rsvd_bytes) {
+		err = -EINVAL;
+		goto err_skip;
+	}
+
+	err = get_exclusive(desc);
+	if (err < 0)
+		goto err_skip;
+
+	err = ubi_start_update(ubi, vol, bytes);
+	if (bytes == 0)
+		revoke_exclusive(desc, UBI_READWRITE);
+err_skip:
+	return err;
+}
+
+ssize_t rootfs_write(struct ubi_volume_desc *desc, const char *buf,
+			      size_t count, int flag)
+{
+	int err = 0;
+	struct ubi_volume *vol = desc->vol;
+	struct ubi_device *ubi = vol->ubi;
+
+	if ( !vol->updating ) {
+		printk("rootfs_write not ready\n");
+		return -1;
+	}
+
+	err = rootfs_ubi_more_update_data(ubi, vol, buf, count, flag);
+
+	if (err < 0) {
+		ubi_err(ubi, "cannot accept more %zd bytes of data, error %d",
+			count, err);
+		return err;
+	}
+
+	if (err) {
+		/*
+		 * The operation is finished, @err contains number of actually
+		 * written bytes.
+		 */
+		count = err;
+
+		if (vol->changing_leb) {
+			revoke_exclusive(desc, UBI_READWRITE);
+			return count;
+		}
+
+		err = ubi_check_volume(ubi, vol->vol_id);
+		if (err < 0)
+			return err;
+
+		if (err) {
+			ubi_warn(ubi, "volume %d on UBI device %d is corrupted",
+				 vol->vol_id, ubi->ubi_num);
+			vol->corrupted = 1;
+		}
+		vol->checked = 1;
+		ubi_volume_notify(ubi, vol, UBI_VOLUME_UPDATED);
+		revoke_exclusive(desc, UBI_READWRITE);
+	}
+
+	return count;
+}
+
+int rootfs_release(struct ubi_volume_desc *desc)
+{
+	struct ubi_volume *vol = desc->vol;
+
+	if (vol->updating) {
+		ubi_warn(vol->ubi, "update of volume %d not finished, volume is damaged",
+			 vol->vol_id);
+		ubi_assert(!vol->changing_leb);
+		vol->updating = 0;
+		vfree(vol->upd_buf);
+	} else if (vol->changing_leb) {
+		dbg_gen("only %lld of %lld bytes received for atomic LEB change for volume %d:%d, cancel",
+			vol->upd_received, vol->upd_bytes, vol->ubi->ubi_num,
+			vol->vol_id);
+		vol->changing_leb = 0;
+		vfree(vol->upd_buf);
+	}
+
+	ubi_close_volume(desc);
+	return 0;
+}
+#endif
