@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2011, Sonos, Inc.
+ * Copyright (c) 2017, Sonos, Inc.
  *
  * SPDX-License-Identifier:	GPL-2.0
  *
@@ -19,9 +19,8 @@
 #include <linux/module.h>
 #include <linux/gpio.h>
 #include <linux/of_gpio.h>
-
-#include "mdp.h"
-extern struct manufacturing_data_page sys_mdp;
+#include <net/netlink.h>
+#include <linux/rtnetlink.h>
 
 #define MV88E6020_REG_PHY_CNTL        0x00
 #define MV88E6020_REG_PHY_SPEC_CNTL   0x10
@@ -109,8 +108,6 @@ MODULE_AUTHOR("Ed Perreault");
 MODULE_LICENSE("GPL");
 
 static struct phy_device *mv88e6020_phydev = NULL;
-static struct net_device *mv88e6020_mac_device = NULL;
-void (*mv88e6020_link_up)(struct net_device *ndev) = NULL;
 static struct work_struct mv88e6020_link_isr_work;
 static struct delayed_work mv88e6020_errata34_work;
 int mv88e6020_irq;
@@ -122,7 +119,7 @@ static unsigned long mv88e6020_errata_count=0;
 
 static struct proc_dir_entry *procdir;
 
-static int swicth_to_cpu_link_status = 0;
+static int switch_to_cpu_link_status = 0;
 
 typedef struct _mv88e6020_env {
 	char	*name;
@@ -139,6 +136,45 @@ static mv88e6020_env_t mv88e6020_named_entries[] = {
 	{ NULL,		0},
 };
 
+void mv88e6020_announce_linkup(struct net_device *dev)
+{
+	struct net *rtnl;
+	struct sk_buff *skb;
+	struct nlmsghdr *nlh;
+	unsigned char *b;
+	int size = NLMSG_SPACE(1024);
+
+	if (dev == NULL) {
+		printk("%s: dev NULL\n", __func__);
+		return;
+	}
+	rtnl = dev_net(dev);
+
+	printk("%s: announcing link status change\n", __func__);
+
+	skb = nlmsg_new(size, GFP_ATOMIC);
+	if (!skb) {
+		netlink_set_err(rtnl->rtnl, 0, RTMGRP_Rincon, ENOBUFS);
+		return;
+	}
+	b = skb->tail;
+	nlh = nlmsg_put(skb, 0, 0, RWM_MII, 0, 0);
+	if ( !nlh )
+		goto nlmsg_failure;
+	nla_put(skb, RWA_DEV_NAME, IFNAMSIZ, dev->name);
+	nlh->nlmsg_len = skb->tail - b;
+
+	NETLINK_CB(skb).dst_group = RTMGRP_Rincon;
+	nlmsg_end(skb, nlh);
+	rtnl_notify(skb, rtnl, 0, RTMGRP_Rincon, NULL, GFP_ATOMIC);
+	return;
+
+nlmsg_failure:
+	kfree_skb(skb);
+	netlink_set_err(rtnl->rtnl, 0, RTMGRP_Rincon, EINVAL);
+	printk("%s: phy announce failed.\n", __func__);
+}
+
 static int mv88e6020_busychk(struct phy_device *phydev)
 {
 	u16 reg = 0;
@@ -372,37 +408,13 @@ static void mv88e6020_check_link(void)
 		}
 	}
 
-	swicth_to_cpu_link_status = link;
+	switch_to_cpu_link_status = link;
 
 	if (mv88e6020_phydev == NULL) {
 		printk("mv88e6020_phydev NULL\n");
+		return;
 	}
-	else if (mv88e6020_link_up != NULL) {
-		(*mv88e6020_link_up)(mv88e6020_mac_device);
-	}
-}
-
-void mv88e6020_sonos_phy_connect(void (*link_up)(struct net_device *), struct net_device *ndev)
-{
-	mv88e6020_link_up = link_up;
-	mv88e6020_mac_device = ndev;
-	mv88e6020_check_link();
-}
-
-void mv88e6020_sonos_phy_disconnect(void)
-{
-	int phy;
-	for (phy = 0; phy <= 1; phy++) {
-		mv88e6020_status[phy].duplex = 0;
-		mv88e6020_status[phy].speed = 0;
-	}
-	if (mv88e6020_link_up != NULL)
-		(*mv88e6020_link_up)(mv88e6020_mac_device);
-	mv88e6020_link_up = NULL;
-	mv88e6020_mac_device = NULL;
-	cancel_delayed_work(&mv88e6020_errata34_work);
-	disable_irq_nosync(mv88e6020_irq);
-	free_irq(mv88e6020_irq, NULL);
+	mv88e6020_announce_linkup(mv88e6020_phydev->attached_dev);
 }
 
 void mv88e6020_sonos_get_port_status(unsigned int port, struct ethtool_cmd *cmd)
@@ -424,13 +436,14 @@ static void mv88e6020_link_change(struct work_struct *work)
 
 		mv88e6020_check_link();
 
-		gb2int = mdiobus_read(mv88e6020_phydev->bus, MV88E6020_GLOBAL2_REGS_ADDR,
-							  MV88E6020_PHY_INTR_STATUS_REG);
+		gb2int = mdiobus_read(mv88e6020_phydev->bus,
+				MV88E6020_GLOBAL2_REGS_ADDR,
+				MV88E6020_PHY_INTR_STATUS_REG);
 		gb2int &= mask;
 
 		if (pass++ > 1000) {
 			printk("%s: exceeded max attempts to clear interrupt\n", __func__);
-		    mv88e6020_max_pass = pass;
+			mv88e6020_max_pass = pass;
 			return;
 		}
 	}
@@ -442,20 +455,21 @@ static void mv88e6020_link_change(struct work_struct *work)
 	}
 }
 
-void mv88e6020_init_irq(void)
+static void mv88e6020_init_irq(struct phy_device *phydev)
 {
 	int ret;
         struct device_node *node = NULL;
 
         node = of_find_compatible_node(NULL, NULL, "Sonos,mv88e6020_phy");
         if (node) {
-                mv88e6020_irq = gpio_to_irq(of_get_named_gpio(node, "interrupts", NULL));
+                mv88e6020_irq = gpio_to_irq(of_get_named_gpio(node, "interrupts", 0));
         }
 
 	if ( mv88e6020_irq < 0 ) {
 		printk("%s: no irq\n", __func__);
 		return;
 	}
+
 	INIT_WORK(&mv88e6020_link_isr_work, mv88e6020_link_change);
 
 	ret = request_irq(mv88e6020_irq, mv88e6020_isr, IRQF_TRIGGER_LOW,
@@ -490,9 +504,9 @@ static void mv88e6020_errata34(struct work_struct *work)
 		}
 	}
 
-	schedule_delayed_work(&mv88e6020_errata34_work, HZ);
 }
 
+#ifdef CONFIG_PROC_FS
 static int mv88e6020_proc_read(struct seq_file *m, void *v)
 {
 	seq_printf(m, "Interupts %ld\n"
@@ -591,11 +605,11 @@ static ssize_t mv88e6020_port_proc_write(struct file *file,
 		peq = strchr(buf, '=');
 		if (peq != NULL) {
 			*peq = 0;
-			if ( strict_strtol(peq+1, 16, &longtemp) != 0 )
+			if ( kstrtol(peq+1, 16, &longtemp) != 0 )
 				return -EIO;
 			val = longtemp;
 			if (strncmp(buf, "reg", 3) == 0) {
-				if ( strict_strtol(buf+3, 16, &longtemp) != 0 )
+				if ( kstrtol(buf+3, 16, &longtemp) != 0 )
 					return -EIO;
 				regnum = longtemp;
 				mdiobus_write(mv88e6020_phydev->bus, devaddr,
@@ -634,11 +648,11 @@ static ssize_t mv88e6020_phy_proc_write(struct file *file,
 		peq = strchr(buf, '=');
 		if (peq != NULL) {
 			*peq = 0;
-			if ( strict_strtol(peq+1, 16, &longtemp) != 0 )
+			if ( kstrtol(peq+1, 16, &longtemp) != 0 )
 				return -EIO;
 			val = longtemp;
 			if (strncmp(buf, "reg", 3) == 0) {
-				if (strict_strtol(buf+3, 16, &longtemp) != 0 )
+				if (kstrtol(buf+3, 16, &longtemp) != 0 )
 					return -EIO;
 				regnum = longtemp;
 				mv88e6020_phy_write(phynum, regnum, val);
@@ -681,6 +695,68 @@ static const struct file_operations mv88e6020_phy_proc_fops = {
 	.write		= mv88e6020_phy_proc_write,
 };
 
+void m88e6020_proc_fs(void)
+{
+	mv88e6020_env_t *entry;
+	int var_num;
+
+	if ( procdir )
+		return;
+
+	procdir = proc_mkdir(MV88E6020_PROC_DIR, NULL);
+	if (procdir == NULL) {
+		printk("Couldn't create base dir /proc/%s\n",
+			MV88E6020_PROC_DIR);
+		return;
+	}
+
+	if (! proc_create_data("data", 0, procdir,
+		&mv88e6020_data_proc_fops, 0)) {
+		printk("mv88e6020/data not created\n");
+		goto cleanup;
+	}
+
+	if ( !proc_create_data("bounce-links", 0, procdir,
+		&mv88e6020_bounce_links_proc_fops, 0)) {
+		printk("mv88e6020/bounce-links not created\n");
+		goto cleanup;
+	}
+
+	entry = mv88e6020_named_entries;
+	while (entry->name && entry->id) {
+		if (!proc_create_data(entry->name, 0644, procdir,
+			&mv88e6020_port_proc_fops, (void *)entry->id))
+			goto cleanup;
+		entry++;
+	}
+
+	for (var_num = 0; var_num <= 1; var_num++) {
+		char name[8];
+		sprintf(name, "phy%d", var_num);
+		if (!proc_create_data(name, 0644, procdir,
+			&mv88e6020_phy_proc_fops, (void *)var_num))
+			 goto cleanup;
+	}
+	return;
+cleanup:
+	remove_proc_subtree(MV88E6020_PROC_DIR, NULL);
+	procdir = NULL;
+	return;
+}
+#endif
+
+#ifdef CONFIG_PM
+static int m88e6020_suspend(struct phy_device *phydev)
+{
+        return 0;
+}
+
+static int m88e6020_resume(struct phy_device *phydev)
+{
+        return 0;
+}
+#endif
+
 static int m88e6020_config_init(struct phy_device *phydev)
 {
 	int phy, err, ident;
@@ -751,7 +827,7 @@ static int m88e6020_config_init(struct phy_device *phydev)
 		printk("%s: write error %d\n", __func__, err);
 	}
 
-	mv88e6020_init_irq();
+	mv88e6020_init_irq(phydev);
 
 	mv88e6020_check_link();
 
@@ -760,7 +836,11 @@ static int m88e6020_config_init(struct phy_device *phydev)
 
 	INIT_DELAYED_WORK(&mv88e6020_errata34_work, mv88e6020_errata34);
 	schedule_delayed_work(&mv88e6020_errata34_work, HZ);
+	mdiobus_write(mv88e6020_phydev->bus, MV88E6020_PORT5_SWITCH_ADDR, 1, 0xC03D);
 
+#ifdef CONFIG_PROC_FS
+	m88e6020_proc_fs();
+#endif
 	return 0;
 }
 
@@ -771,7 +851,7 @@ static int m88e6020_config_aneg(struct phy_device *phydev)
 
 static int m88e6020_genphy_read_status(struct phy_device *phydev)
 {
-	if ( swicth_to_cpu_link_status ) {
+	if ( switch_to_cpu_link_status ) {
 		phydev->speed = SPEED_100;
 		phydev->duplex = DUPLEX_FULL;
 		phydev->link = 1;
@@ -795,16 +875,20 @@ static int m88e6020_config_intr(struct phy_device *phydev)
 
 static struct phy_driver marvell_drivers[] = {
 	{
-		.phy_id = 0x00000200,
-		.phy_id_mask = 0x000fff00,
-		.name = "Marvell 88E6020",
-		.features = PHY_GBIT_FEATURES,
-		.flags = PHY_HAS_INTERRUPT,
-		.config_init = &m88e6020_config_init,
-		.config_aneg = &m88e6020_config_aneg,
-		.read_status = &m88e6020_genphy_read_status,
+		.phy_id        = 0x00000200,
+		.phy_id_mask   = 0x000fff00,
+		.name          = "Marvell 88E6020",
+		.features      = PHY_BASIC_FEATURES,
+		.flags         = PHY_HAS_INTERRUPT,
+		.config_init   = &m88e6020_config_init,
+		.config_aneg   = &m88e6020_config_aneg,
+		.read_status   = &m88e6020_genphy_read_status,
 		.ack_interrupt = &m88e6020_ack_interrupt,
-		.config_intr = &m88e6020_config_intr,
+		.config_intr   = &m88e6020_config_intr,
+#ifdef CONFIG_PM
+		.suspend       = &m88e6020_suspend,
+		.resume        = &m88e6020_resume,
+#endif
 		.driver = { .owner = THIS_MODULE },
 	},
 };
@@ -812,90 +896,33 @@ static struct phy_driver marvell_drivers[] = {
 static int __init mv88e6020_init(void)
 {
 	int ret = 0;
-	mv88e6020_env_t *entry;
-	int var_num;
-
-#ifdef CONFIG_SONOS_SOLBASE
-	if ( sys_mdp.mdp_revision >= MDP_REVISION_SOLBASE_PROTO4 ) {
-		return 0;
-	}
-#endif
 
+	switch_to_cpu_link_status = 0;
 	mv88e6020_phydev = NULL;
+	procdir = NULL;
+
 	ret = phy_drivers_register(marvell_drivers,
 		ARRAY_SIZE(marvell_drivers));
 
-	if ( ret )
-		goto err;
-
-	swicth_to_cpu_link_status = 0;
-	procdir = proc_mkdir(MV88E6020_PROC_DIR, NULL);
-	if (procdir == NULL) {
-		printk("Couldn't create base dir /proc/%s\n",
-			MV88E6020_PROC_DIR);
-		phy_drivers_unregister(marvell_drivers,
-			ARRAY_SIZE(marvell_drivers));
-		return -ENOMEM;
-	}
-
-	if (! proc_create_data("data", 0, procdir,
-		&mv88e6020_data_proc_fops, 0)) {
-		printk("mv88e6020/data not created\n");
-		goto cleanup;
-	}
-
-	if ( !proc_create_data("bounce-links", 0, procdir,
-		&mv88e6020_bounce_links_proc_fops, 0)) {
-		printk("mv88e6020/bounce-links not created\n");
-		goto cleanup;
-	}
-
-	entry = mv88e6020_named_entries;
-	while (entry->name && entry->id) {
-		if (!proc_create_data(entry->name, 0644, procdir,
-			&mv88e6020_port_proc_fops, (void *)entry->id))
-			goto cleanup;
-		entry++;
-	}
-
-	for (var_num = 0; var_num <= 1; var_num++) {
-		char name[8];
-		sprintf(name, "phy%d", var_num);
-		if (!proc_create_data(name, 0644, procdir,
-			&mv88e6020_phy_proc_fops, (void *)var_num))
-			 goto cleanup;
-	}
-	return 0;
-cleanup:
-	phy_drivers_unregister(marvell_drivers,
-			ARRAY_SIZE(marvell_drivers));
-	remove_proc_subtree(MV88E6020_PROC_DIR, NULL);
-err:
 	return ret;
 }
 
 static void __exit mv88e6020_exit(void)
 {
-	printk("%s: done\n", __func__);
-#ifdef CONFIG_SONOS_SOLBASE
-	if ( sys_mdp.mdp_revision >= MDP_REVISION_SOLBASE_PROTO4 ) {
-		return;
-	}
-#endif
+#ifdef CONFIG_PROC_FS
 	remove_proc_subtree(MV88E6020_PROC_DIR, NULL);
+#endif
 	phy_drivers_unregister(marvell_drivers,
 			ARRAY_SIZE(marvell_drivers));
 	mv88e6020_phydev = NULL;
 }
 
-#if defined(CONFIG_SONOS_PARAMOUNT) || defined(CONFIG_SONOS_NEPTUNE)
-int phy_power_down(struct phy_device *phydev, int on)
+int mv88e6020_phy_power_down(struct phy_device *phydev, int on)
 {
     return 0;
 }
-#endif
 
-module_init(mv88e6020_init);
+subsys_initcall(mv88e6020_init);
 module_exit(mv88e6020_exit);
 
 static struct mdio_device_id __maybe_unused marvell_tbl[] = {
