@@ -22,7 +22,7 @@ typedef struct {
         size_t keyIdLen;
 } SonosRsaKeyEntry;
 
-extern const SonosRsaKeyEntry * const g_SonosSigningKeys[];
+extern const SonosRsaKeyEntry * g_SonosSigningKeys[];
 
 extern void sonosInitKeyTable(void);
 
