@@ -132,6 +132,13 @@ enum {
 	RTM_GETMDB = 86,
 #define RTM_GETMDB RTM_GETMDB
 
+	RTM_NEWNSID = 88,
+#define RTM_NEWNSID RTM_NEWNSID
+	RTM_DELNSID = 89,
+#define RTM_DELNSID RTM_DELNSID
+	RTM_GETNSID = 90,
+#define RTM_GETNSID RTM_GETNSID
+
 	__RTM_MAX,
 #define RTM_MAX		(((__RTM_MAX + 3) & ~3) - 1)
 };
@@ -153,7 +160,7 @@ struct rtattr {
 
 /* Macros to handle rtattributes */
 
-#define RTA_ALIGNTO	4
+#define RTA_ALIGNTO	4U
 #define RTA_ALIGN(len) ( ((len)+RTA_ALIGNTO-1) & ~(RTA_ALIGNTO-1) )
 #define RTA_OK(rta,len) ((len) >= (int)sizeof(struct rtattr) && \
 			 (rta)->rta_len >= sizeof(struct rtattr) && \
@@ -235,6 +242,7 @@ enum {
 #define RTPROT_NTK	15	/* Netsukuku */
 #define RTPROT_DHCP	16      /* DHCP client */
 #define RTPROT_MROUTED	17      /* Multicast daemon */
+#define RTPROT_BABEL	42      /* Babel daemon */
 
 /* rtm_scope
 
@@ -262,6 +270,7 @@ enum rt_scope_t {
 #define RTM_F_CLONED		0x200	/* This route is cloned		*/
 #define RTM_F_EQUALIZE		0x400	/* Multipath equalizer: NI	*/
 #define RTM_F_PREFIX		0x800	/* Prefix addresses		*/
+#define RTM_F_LOOKUP_TABLE	0x1000	/* set rtm_table to FIB lookup result */
 
 /* Reserved table identifiers */
 
@@ -297,6 +306,12 @@ enum rtattr_type_t {
 	RTA_TABLE,
 	RTA_MARK,
 	RTA_MFC_STATS,
+	RTA_VIA,
+	RTA_NEWDST,
+	RTA_PREF,
+	RTA_ENCAP_TYPE,
+	RTA_ENCAP,
+	RTA_UID,
 	__RTA_MAX
 };
 
@@ -326,6 +341,10 @@ struct rtnexthop {
 #define RTNH_F_DEAD		1	/* Nexthop is dead (used by multipath)	*/
 #define RTNH_F_PERVASIVE	2	/* Do recursive gateway lookup	*/
 #define RTNH_F_ONLINK		4	/* Gateway is forced on link	*/
+#define RTNH_F_OFFLOAD		8	/* offloaded route */
+#define RTNH_F_LINKDOWN		16	/* carrier-down on nexthop */
+
+#define RTNH_COMPARE_MASK	(RTNH_F_DEAD | RTNH_F_LINKDOWN)
 
 /* Macros to handle hexthops */
 
@@ -338,6 +357,12 @@ struct rtnexthop {
 #define RTNH_SPACE(len)	RTNH_ALIGN(RTNH_LENGTH(len))
 #define RTNH_DATA(rtnh)   ((struct rtattr*)(((char*)(rtnh)) + RTNH_LENGTH(0)))
 
+/* RTA_VIA */
+struct rtvia {
+	__kernel_sa_family_t	rtvia_family;
+	__u8			rtvia_addr[0];
+};
+
 /* RTM_CACHEINFO */
 
 struct rta_cacheinfo {
@@ -386,15 +411,22 @@ enum {
 #define RTAX_RTO_MIN RTAX_RTO_MIN
 	RTAX_INITRWND,
 #define RTAX_INITRWND RTAX_INITRWND
+	RTAX_QUICKACK,
+#define RTAX_QUICKACK RTAX_QUICKACK
+	RTAX_CC_ALGO,
+#define RTAX_CC_ALGO RTAX_CC_ALGO
 	__RTAX_MAX
 };
 
 #define RTAX_MAX (__RTAX_MAX - 1)
 
-#define RTAX_FEATURE_ECN	0x00000001
-#define RTAX_FEATURE_SACK	0x00000002
-#define RTAX_FEATURE_TIMESTAMP	0x00000004
-#define RTAX_FEATURE_ALLFRAG	0x00000008
+#define RTAX_FEATURE_ECN	(1 << 0)
+#define RTAX_FEATURE_SACK	(1 << 1)
+#define RTAX_FEATURE_TIMESTAMP	(1 << 2)
+#define RTAX_FEATURE_ALLFRAG	(1 << 3)
+
+#define RTAX_FEATURE_MASK	(RTAX_FEATURE_ECN | RTAX_FEATURE_SACK | \
+				 RTAX_FEATURE_TIMESTAMP | RTAX_FEATURE_ALLFRAG)
 
 struct rta_session {
 	__u8	proto;
@@ -513,48 +545,48 @@ enum {
 #define TCA_PAYLOAD(n) NLMSG_PAYLOAD(n,sizeof(struct tcmsg))
 
 /**************************************************************
- *     Rincon Wireless Messages and Attributes
+ *       Rincon Wireless Messages and Attributes
  ****/
 
 enum
 {
-    RWM_ASSOC,
-    RWM_PROBE,
-    RWM_HHID,
-    RWM_MII,
-    RWM_INFRA_NEIGHBOR,
-    RWM_NETSTART_STA,
-    RWM_CLIENT_ASSOC,
-    RWM_SUPPLICANT,
-    RWM_SONOSNET_PATH,
-    RWM_RENEW_DHCP,
-    RWM_SONOSNET_FULL_MESH,
-    RWM_PRIMARY_FOUND,
-    RWM_PRIMARY_LOST,
-    RWM_DUPE_IP,
-    RWM_HTAP_CONN,
-    RWM_CMD_TIMEOUT,
-    RWM_NETWORK_SETTINGS,
+	RWM_ASSOC,
+	RWM_PROBE,
+	RWM_HHID,
+	RWM_MII,
+	RWM_INFRA_NEIGHBOR,
+	RWM_NETSTART_STA,
+	RWM_CLIENT_ASSOC,
+	RWM_SUPPLICANT,
+	RWM_SONOSNET_PATH,
+	RWM_RENEW_DHCP,
+	RWM_SONOSNET_FULL_MESH,
+	RWM_PRIMARY_FOUND,
+	RWM_PRIMARY_LOST,
+	RWM_DUPE_IP,
+	RWM_HTAP_CONN,
+	RWM_CMD_TIMEOUT,
+	RWM_NETWORK_SETTINGS,
 };
 
 enum
 {
-    RWA_DEV_NAME,
-    RWA_ASSOC_MAC,
-    RWA_ASSOC_UUID,
-    RWA_PROBE_MAC,
-    RWA_PROBE_DATA,
-    RWA_HHID,
-    RWA_INFRA_NEIGHBOR_MAC,
-    RWA_RSSI,
-    RWA_RSSI_PER_CHAIN,
-    RWA_NETSTART_STA_MAC,
-    RWA_CLIENT_ASSOC_DATA,
-    RWA_SUPPLICANT_EVENT,
-    RWA_SUPPLICANT_EVENT_DATA,
-    RWA_PATH_LENGTH,
-    RWA_DUPE_TIME,
-    RWA_NETWORK_SETTINGS_DATA,
+	RWA_DEV_NAME,
+	RWA_ASSOC_MAC,
+	RWA_ASSOC_UUID,
+	RWA_PROBE_MAC,
+	RWA_PROBE_DATA,
+	RWA_HHID,
+	RWA_INFRA_NEIGHBOR_MAC,
+	RWA_RSSI,
+	RWA_RSSI_PER_CHAIN,
+	RWA_NETSTART_STA_MAC,
+	RWA_CLIENT_ASSOC_DATA,
+	RWA_SUPPLICANT_EVENT,
+	RWA_SUPPLICANT_EVENT_DATA,
+	RWA_PATH_LENGTH,
+	RWA_DUPE_TIME,
+	RWA_NETWORK_SETTINGS_DATA,
 };
 
 /********************************************************************
@@ -658,7 +690,11 @@ enum rtnetlink_groups {
 #define RTNLGRP_IPV6_NETCONF	RTNLGRP_IPV6_NETCONF
 	RTNLGRP_MDB,
 #define RTNLGRP_MDB		RTNLGRP_MDB
-	RTNLGRP_RINCON,
+	RTNLGRP_MPLS_ROUTE,
+#define RTNLGRP_MPLS_ROUTE	RTNLGRP_MPLS_ROUTE
+	RTNLGRP_NSID,
+#define RTNLGRP_NSID		RTNLGRP_NSID
+        RTNLGRP_RINCON,
 	__RTNLGRP_MAX
 };
 #define RTNLGRP_MAX	(__RTNLGRP_MAX - 1)
@@ -682,6 +718,8 @@ struct tcamsg {
 /* New extended info filters for IFLA_EXT_MASK */
 #define RTEXT_FILTER_VF		(1 << 0)
 #define RTEXT_FILTER_BRVLAN	(1 << 1)
+#define RTEXT_FILTER_BRVLAN_COMPRESSED	(1 << 2)
+#define	RTEXT_FILTER_SKIP_STATS	(1 << 3)
 
 /* End of information exported to user level */
 
