@@ -27,6 +27,7 @@
 #include <linux/export.h>
 #include <linux/utsname.h>
 #include <linux/sched.h>
+#include <linux/reboot.h>
 #include <asm/unaligned.h>
 
 #include <net/bluetooth/bluetooth.h>
@@ -43,6 +44,8 @@ static DEFINE_IDA(sock_cookie_ida);
 
 static atomic_t monitor_promisc = ATOMIC_INIT(0);
 
+static int hci_system_rebooting = 0;
+
 /* ----- HCI socket interface ----- */
 
 /* Socket info */
@@ -517,7 +520,7 @@ static struct sk_buff *create_monitor_ctrl_open(struct sock *sk)
 	put_unaligned_le16(format, skb_put(skb, 2));
 	memcpy(skb_put(skb, sizeof(ver)), ver, sizeof(ver));
 	put_unaligned_le32(flags, skb_put(skb, 4));
-	*(ushort *)skb_put(skb, 1) = TASK_COMM_LEN;
+	*skb_put(skb, 1) = TASK_COMM_LEN;
 	memcpy(skb_put(skb, TASK_COMM_LEN), hci_pi(sk)->comm, TASK_COMM_LEN);
 
 	__net_timestamp(skb);
@@ -616,7 +619,7 @@ send_monitor_note(struct sock *sk, const char *fmt, ...)
 
 	va_start(args, fmt);
 	vsprintf(skb_put(skb, len), fmt, args);
-	*(ushort*)skb_put(skb, 1) = 0;
+	*skb_put(skb, 1) = 0;
 	va_end(args);
 
 	__net_timestamp(skb);
@@ -1402,7 +1405,7 @@ static void hci_sock_cmsg(struct sock *sk, struct msghdr *msg,
 	}
 }
 
-static int hci_sock_recvmsg(struct kiocb *iocb, struct socket *sock, struct msghdr *msg,
+static int hci_sock_recvmsg(struct socket *sock, struct msghdr *msg,
 			    size_t len, int flags)
 {
 	int noblock = flags & MSG_DONTWAIT;
@@ -1666,7 +1669,7 @@ drop:
 	return err;
 }
 
-static int hci_sock_sendmsg(struct kiocb *iocb, struct socket *sock, struct msghdr *msg,
+static int hci_sock_sendmsg(struct socket *sock, struct msghdr *msg,
 			    size_t len)
 {
 	struct sock *sk = sock->sk;
@@ -1677,10 +1680,14 @@ static int hci_sock_sendmsg(struct kiocb *iocb, struct socket *sock, struct msgh
 
 	BT_DBG("sock %p sk %p", sock, sk);
 
+	if (hci_system_rebooting) {
+		hci_sock_clear_flag(sk, HCI_UP);
+	}
+
 	if (msg->msg_flags & MSG_OOB)
 		return -EOPNOTSUPP;
 
-	if (msg->msg_flags & ~(MSG_DONTWAIT|MSG_NOSIGNAL|MSG_ERRQUEUE|MSG_CMSG_COMPAT))
+	if (msg->msg_flags & ~(MSG_DONTWAIT|MSG_NOSIGNAL|MSG_ERRQUEUE))
 		return -EINVAL;
 
 	if (len < 4 || len > HCI_MAX_FRAME_SIZE)
@@ -2021,6 +2028,18 @@ static const struct net_proto_family hci_sock_family_ops = {
 	.create	= hci_sock_create,
 };
 
+static int hci_sock_reboot_event(struct notifier_block *this,
+				unsigned long event, void *ptr)
+{
+	hci_system_rebooting = 1;
+
+	return NOTIFY_DONE;
+}
+
+static struct notifier_block hci_sock_reboot_notifier = {
+	.notifier_call = hci_sock_reboot_event,
+};
+
 int __init hci_sock_init(void)
 {
 	int err;
@@ -2044,6 +2063,8 @@ int __init hci_sock_init(void)
 		goto error;
 	}
 
+	err = register_reboot_notifier(&hci_sock_reboot_notifier);
+
 	BT_INFO("HCI socket layer initialized");
 
 	return 0;
